import bisect
import itertools
from typing import Dict, List, Tuple

from components import Component
from node import Node, Protein, Peptide


class Graph:
    """
    The node_to_delete, discovered_nodes, explored_nodes are dict
    with key as the first id of the node, and an empty string as value
    """
    node_dict: Dict[Node, List[Node]]
    node_to_delete: Dict[str, str]
    accession_object: Dict[str, Node]
    discovered_nodes: Dict[str, str]
    explored_nodes: Dict[str, str]
    sorted_keys = List[Node]
    sorted_keys_id = List[str]

    def __init__(self):
        self.node_dict = {}
        self.node_to_delete = {}
        self.accession_object = {}
        self.discovered_nodes = {}
        self.explored_nodes = {}

    def get_node_dict_keys(self):
        return self.node_dict.keys()

    def get_node_dict(self):
        return self.node_dict

    """methods for step 1: initialize"""

    def add_peptide(self,
                    peptide_id_list: List[Tuple[str, float, int]]) -> None:

        # for every peptide and score pair that is a hit
        for peptide_id, peptide_score, decoy in peptide_id_list:
            # make a Peptide node
            current_peptide = Peptide([peptide_id], peptide_score, decoy)
            # store that peptide into the graph
            self.node_dict[current_peptide] = []

    def add_protein(self, protein_id_list: List[Tuple[str, int]],
                    linked_peptide_dict: Dict[str, List[Tuple[str, float, int]]],
                    protein_accession_dict: Dict[str, List[str]]) \
            -> None:
        """
        the protein with this accession may already be in the graph
        because in the sqlite database, 1 protein sqlite id correspond to
        multiple protein accessions, and they may overlap (a protein accession
        can exist in multiple protein sqlite ids)
        """
        for protein_id, decoy in protein_id_list:

            protein_accession_list = protein_accession_dict[protein_id]
            for protein_accession in protein_accession_list:
                current_protein = Protein([protein_accession], protein_id,
                                          decoy)

                # if node is not in graph, make empty list first
                if not self.node_in_graph(current_protein):
                    self.node_dict[current_protein] = []
                self.make_edge_from_protein_id(current_protein,
                                               linked_peptide_dict)

            # TODO testing
            if 'sp|Q9UI08-5|EVL_HUMAN' in protein_accession_list:
                print('what', protein_id, decoy)


    def make_edge_from_protein_id(self,
                                  the_protein: Protein,
                                  linked_peptide_dict:
                                  Dict[str, List[
                                      Tuple[str, float, int]]]) \
            -> None:
        """
        based the protein given, add peptides as edges to the current protein
        """

        protein_id = the_protein.get_first_sqlite_id()

        if protein_id not in linked_peptide_dict:
            return

        neighbour_list_pep = linked_peptide_dict[protein_id]
        # neighbour list is just a list of peptide ids, with score and decoy
        at_least_one = False
        for neighbour_id, neighbour_score, decoy in neighbour_list_pep:
            a_peptide = Peptide([neighbour_id], neighbour_score, decoy)
            if self.node_in_graph(a_peptide):
                self.node_dict[the_protein].append(a_peptide)
                at_least_one = True
        if not at_least_one:
            self.delete_node(the_protein)

    def make_edges_from_peptide(self,
                                linked_protein_dict: Dict[str, List[Tuple[str, int]]],
                                protein_accession_dict: Dict[str, List[str]]) \
            -> None:
        """
        iterate through the peptide keys, and for each, get their link (most
        times there is only one link), for each link, get all the accessions.
        for each accession, make a protein object, check if is a protein key,
        then add it to this peptide keys' value
        """
        for node in self.node_dict:
            if isinstance(node, Peptide):

                peptide_id = node.get_first_id()

                if peptide_id not in linked_protein_dict:
                    continue

                # a list of protein sqlite id that links to this peptide
                neighbour_list_pro = linked_protein_dict[peptide_id]
                for protein_id, decoy in neighbour_list_pro:
                    accession_list = protein_accession_dict[protein_id]
                    for accession in accession_list:
                        a_protein = Protein([accession], protein_id, decoy)
                        self.node_dict[node].append(a_protein)

    def node_in_graph(self, a_node: Node):
        """
        check if this node is in the graph
        :param a_node: this node to be checked
        :return: whether the node is in the graph (boolean)
        """
        if self.get_node_dict().get(a_node) is None:
            in_graph = False
        else:
            in_graph = True

        return in_graph

    def store(self) -> None:
        pass
        # with open('initialized_graph.csv', 'w', newline='\n') as csvfile:
        #     fieldnames = []
        #     writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        #     writer.writeheader()
        #     writer.writerows(self.node_dict)

    def get_sort_keys(self) -> None:

        """
        :param
        """

        # convert dict key to list (since dict key cannot be sorted)
        # sort, then store
        keys_list = list(self.node_dict.keys())
        keys_list.sort()
        sorted_keys_list = keys_list
        self.sorted_keys = sorted_keys_list

        # get all id (first id + target_decoy) of those keys (in sorted order)
        sorted_id_list = []
        for keys in sorted_keys_list:
            sorted_id_list.append(keys.get_first_id() + keys.get_target_decoy())

        # sorted_id_list is then in the same order as sorted keys
        self.sorted_keys_id = sorted_id_list

    """methods for step 2: collapse"""

    def collapse_graph(self) -> None:

        # take all nodes (keys)
        key_list = list(self.node_dict)

        # then group them based on their degree as a list of lists
        all_nodes_reorganized = self.reorder_neighbours(key_list)

        # TODO: Better name
        max_degree = len(all_nodes_reorganized)

        # the value is n when the nodes from degree n to 1 are finished
        progress_count_in_degrees = 1

        # i.e. for every degree
        for num_edges in range(len(all_nodes_reorganized)):

            # skip the group of protein that only map 0 proteins
            # i.e. skip the degree of zero
            if num_edges == 0:
                continue

            # these are all the nodes with the same degree
            same_edges_node_list = all_nodes_reorganized[num_edges]

            # if there are no node that have exactly num_edges degree
            if len(same_edges_node_list) == 0:
                print(progress_count_in_degrees, max_degree, "no node")
                progress_count_in_degrees += 1
                continue

            # start this by first subset (and clone) the node dict
            specific_dict = {}
            self.build_specific_node_dict(same_edges_node_list, specific_dict)
            # then pass the neighbour list and the dict into recursion
            # first protein to remove is key, since all neighbours in
            # neighbour_list share this
            same_edges_node_list.sort()
            self.grouping_recursion(same_edges_node_list, specific_dict)

            print(progress_count_in_degrees, max_degree)
            progress_count_in_degrees += 1

    def collapse_graph_old(self) -> None:
        # for each node
        progress_count = 1
        for key, value in self.node_dict.items():

            # if the node has been delete, skip
            if key.get_first_id() in self.node_to_delete:
                progress_count += 1
                continue

            # code for skipping proteins
            # if isinstance(key, Protein):
            #     progress_count += 1
            #     continue

            # current neighbours: neighbours of the current node
            current_neighbours = value

            # if there are more than 2 neighbours
            if len(current_neighbours) >= 2:

                # actually reorganize the neighbours
                neighbours_reorganized = self.reorder_neighbours(
                    current_neighbours)

                # print(list(map(len, neighbours_reorganized)))

                # the other method of merging
                # not using anymore since it is a lot slower
                # it was used to check mergeability directly
                # if isinstance(key, Peptide):
                #     self.check_for_merging(neighbours_reorganized)
                #     continue

                # use recursion to group protein and then merge them
                # TODO, change to range (1, len(...))?
                for num_protein_mapped_to in range(len(neighbours_reorganized)):
                    # skip the first one, since that peptide in that neighbour
                    # list maps to exactly 0 protein
                    if num_protein_mapped_to == 0:
                        continue

                    neighbours_list = neighbours_reorganized[
                        num_protein_mapped_to]

                    # there are no peptide that map to num_protein_mapped_to protein
                    if len(neighbours_list) == 0:
                        continue

                    # specific_dict is just a subset (and clone) of node dict
                    specific_dict = {}
                    self.build_specific_node_dict(neighbours_list,
                                                  specific_dict)
                    # then pass the neighbour list and the dict into recursion
                    # first protein to remove is key, since all neighbours in
                    # neighbour_list share this
                    self.grouping_recursion(neighbours_list, specific_dict)

            progress_count += 1
            print("nodes collapsed", progress_count, len(self.node_dict),
                  "node is peptide", isinstance(key, Peptide))

    """methods for step 2a: reordering"""

    def find_node_most_edges(self, node_list: List[Node]) -> int:
        max_edges = 0

        for current_node in node_list:
            neighbour_list = self.node_dict[current_node]
            if len(neighbour_list) > max_edges:
                max_edges = len(neighbour_list)

        return max_edges

    def num_neighbour(self, node: Node) -> int:
        return len(self.node_dict.get(node))

    def reorder_neighbours(self, current_neighbours: List[Node]) \
            -> List[List[Node]]:
        """
        take the list of protein, current_neighbours and reformat it into a
        list of list of protein where every protein in each sublist has the same
        number of neighbours, which is returned
        :param current_neighbours:
        :return: neighbours_reorganized [0, 0, 28] means of the current
        neighbours of this protein (which are peptide). 0 map to exactly 0
        protein, 0 map to exactly 1 protein, 28 map to exactly 2 proteins
        """

        neighbours_reorganized = []

        num_most_edges = self.find_node_most_edges(current_neighbours)

        # append as many list as num_most_edges
        # where num_most_edges is the greatest number of
        # neighbours a protein has
        for i in range(0, num_most_edges + 1):
            neighbours_reorganized.append([])

        # for all neighbour
        for current_protein in current_neighbours:
            # find num of neighbours for current node
            num_peptide_neighbours = self.num_neighbour(current_protein)
            # append them to the list of list appropriately
            neighbours_reorganized[
                num_peptide_neighbours].append(current_protein)

        return neighbours_reorganized

    def build_specific_node_dict(self,
                                 same_degree_protein: List[Node],
                                 protein_peptide_dict: Dict[Node, List[Node]]) \
            -> None:
        """
        terminology assumes we are trying to merge peptides
        :param protein_peptide_dict:
        :param same_degree_protein is all the peptides that have the same number
        of protein that they map to
        :return type should be peptide map to a list of protein since I am only
        using this for collapsing peptides
        """

        for peptide in same_degree_protein:
            # is a neighbour and a protein
            neighbour_and_protein = self.node_dict.get(peptide, [])
            if len(neighbour_and_protein) == 0:
                print('something is wrong')
                exit('something is wrong')
            protein_peptide_dict[peptide] = sorted(neighbour_and_protein)

    def grouping_recursion(self, same_degree_protein: List[Node],
                           protein_peptide_dict: Dict[Node, List[Node]]):
        """terminology assume we are merging protein
        first build a dict based on the protein given, but without the
        protein that is these peptides have in common for sure (this is done by
        build_specific_node_dict method)

        second check how many protein do each of the peptides have left
        this should all be the same number since the initial peptide inputted
        map to the same number of protein

        Third if there are no protein left, then merge them
        else group them by their current first protein
        and pass each group with the current first protein
        into recursion
        """

        num_peptide = []

        # for each protein with the same degree
        for protein in same_degree_protein:
            # get its neighbouring peptides
            protein_list = protein_peptide_dict.get(protein)
            # and count
            num_peptide.append(len(protein_list))

        # the length of the set is the number of unique element in num_peptide (
        # i.e. how many peptides do these same_degree_protein connect to
        # which should be 1, meaning that they all have the same degree
        assert len(set(num_peptide)) < 2

        # this is the number of protein that not yet been used for grouping
        # it is initially the degree
        num_protein_all = list(set(num_peptide))[0]


        # TODO: test
        all_id = []
        for protein in same_degree_protein:
            accession = protein.get_first_id()
            if not 'DECOY' in accession and not accession.isnumeric():
                all_id.append(protein.get_first_id())

        if 'sp|Q9UI08-5|EVL_HUMAN' in all_id:
            print('before merging', print(num_protein_all))
            print('all id', all_id)

        # the protein in same_degree_protein all were part of a list that
        # all map to the same set of peptide

        # base case (we have gotten to final protein)
        # if there are no other peptide that it map to, then this means
        # this protein only map to this set
        # this means that all the protein with the same degree has been grouped
        # into groups where within each group all the protein's peptides are the same
        # each function call that end up in the base case only is given a set of
        # protein that share all its peptide
        if num_protein_all == 0:

            # TODO when it was sort() there was nothing wrong (problem)
            sorted_same_degree_protein = sorted(same_degree_protein)
            # merge the protein that share the same set of peptides
            # TODO: refactor and extract this as a method
            # choose the first object in the list as the one to merge to
            final_protein_group = sorted_same_degree_protein[0]

            # TODO: remove test
            if 'sp|Q9UI08-2|EVL_HUMAN' in final_protein_group.get_id():
                print('protein of interest', final_protein_group.get_id())
                print('interest first id', final_protein_group.get_first_id())
                for protein in sorted_same_degree_protein:
                    print('other protein', protein.get_id())
                    print('other protein first id', protein.get_first_id())

            # for all the protein in this set, skipping first one
            for index in range(1, len(sorted_same_degree_protein)):
                protein = sorted_same_degree_protein[index]
                self.delete_node(protein)

                id_list = protein.get_id()
                score = protein.get_score()

                final_protein_group.add_id(id_list)
                # this only keeps the max score of them two
                final_protein_group.add_score(score)

            if 'sp|Q9UI08-2|EVL_HUMAN' in final_protein_group.get_id():
                print('final protein group of interest', final_protein_group.get_id())

            # TODO supposed bottleneck
            # final_id_list = final_protein_group.get_id()
            # self.key_add_id(final_protein_group, final_id_list)

        # recursive case, not 0
        else:
            protein_same_next_peptide_dict = {}

            # for every entry in this group of protein
            for protein in same_degree_protein:

                # get that protein
                peptide = protein_peptide_dict.get(protein, [])

                assert peptide != []

                # sort their peptides
                sorted_peptide = sorted(peptide)
                next_same_peptide = sorted_peptide[0]

                # group the peptide by their next same protein
                protein_same_next_peptide_dict \
                    .setdefault(next_same_peptide, []).append(protein)

            # now in protein_same_next_peptide_dict, protein in each sublist
            # have the same first peptide

            # recursion
            # each protein list share the first peptide on the list
            for next_same_peptide, protein_list in protein_same_next_peptide_dict.items():
                for protein in protein_list:
                    # get each protein
                    protein = protein_peptide_dict.get(protein, [])
                    assert protein != []
                    # remove that protein
                    protein.remove(next_same_peptide)

                protein_list.sort()
                self.grouping_recursion(protein_list, protein_peptide_dict)

    """methods for step 2b: merging"""

    def check_for_merging_old(self, neighbours_reorganized: List[List[Node]]) \
            -> None:
        # all the protein neighbours in the same protein_neighbour_list
        # should have the same length
        for neighbour_list in neighbours_reorganized:
            for neighbour_pair in itertools.combinations(neighbour_list, 2):
                if (
                        # only if the 2 both neighbour have not been deleted yet
                        neighbour_pair[0].get_first_id() not in self.node_to_delete
                        and
                        neighbour_pair[1].get_first_id() not in self.node_to_delete
                        # then compare their neighbours
                ) and self.compare_neighbours_old(neighbour_pair):

                    # either both target and both decoy
                    assert neighbour_pair[0].get_target_decoy() == \
                           neighbour_pair[1].get_target_decoy()

                    self.delete_node(neighbour_pair[1])
                    id_list = neighbour_pair[1].get_id()
                    score = neighbour_pair[1].get_score()
                    neighbour_pair[0].add_id(id_list)
                    # this only keeps the max score of them two
                    neighbour_pair[0].add_score(score)
                    # TODO: bottleneck
                    self.key_add_id(neighbour_pair[0], id_list)

    def compare_neighbours_old(self, node_pair: Tuple[Node, ...]) \
            -> bool:
        """
        if either one is None, return false
        :param node_pair: the pair of node to be compared
        :return: True if the neighbour of the node pair are the same
        """
        # if their first id and target+decoy is the same, then it is the same
        this_node = node_pair[0]
        that_node = node_pair[1]
        these_neighbours = self.node_dict.get(this_node)
        those_neighbours = self.node_dict.get(that_node)
        if these_neighbours is None or those_neighbours is None:
            return False
        else:
            these_neighbours.sort()
            those_neighbours.sort()

        assert len(these_neighbours) == len(those_neighbours)

        # I don't know if python ordered list comparison for equality has
        # early-stop, such that when it reaches an index that is different
        # it exits the equality comparison

        is_the_same = True
        for node_id in range(len(these_neighbours)):
            if these_neighbours[node_id] != those_neighbours[node_id]:
                is_the_same = False
                break

        return is_the_same

    def delete_node(self, current_node: Node) -> None:

        # store the id of the node into the "node_to_delete" dict
        # with an empty string
        self.node_to_delete[current_node.get_first_id()
                            + current_node.get_target_decoy()] = ''

    def key_add_id(self, same_protein: Node, accession_list) -> None:

        # find this protein in the id_list, with binary search
        protein_id = same_protein.get_first_id() + same_protein.get_target_decoy()
        if protein_id in self.sorted_keys_id:
            # find the index of the node in sorted keys, based on its
            # (first id + target/decoy)
            index = bisect.bisect_left(self.sorted_keys_id, protein_id)

            # it is the same index for the protein object in the sorted_key_list
            # so then access it from sorted keys using list indexing
            # this gets me the protein
            equivalent_protein = self.sorted_keys[index]
            equivalent_protein.add_id(accession_list)
            equivalent_protein.add_score(same_protein.get_score())

    """methods for step 3: separate"""

    def make_accession_object_dict(self) -> None:

        # for all node
        for node in self.node_dict:
            # if not deleted
            if (node.get_first_id() + node.get_target_decoy()) not in self.node_to_delete:
                # map its first id to the node object itself
                self.accession_object[node.get_first_id()] = node

    def set_discovered(self, start_node: Node) -> None:
        self.discovered_nodes[start_node.get_first_id()
                              + start_node.get_target_decoy()] = ''

    def set_explored(self, start_node: Node) -> None:
        self.explored_nodes[start_node.get_first_id()
                            + start_node.get_target_decoy()] = ''

    def is_white(self, node: Node) -> bool:
        return (node.get_first_id() + node.get_target_decoy()
                not in self.discovered_nodes) \
               and (node.get_first_id() + node.get_target_decoy()
                    not in self.explored_nodes)

    def is_discovered(self, node) -> bool:
        return node.get_first_id() in self.discovered_nodes

    def is_explored(self, node) -> bool:
        return node.get_first_id() in self.explored_nodes

    def dfs(self, start_node: Node, a_component: Component) -> None:

        self.set_discovered(start_node)

        # if protein, get a peptide list, if peptide get a protein list
        neighbour_list = self.node_dict.get(start_node)

        # for all neighbouring white node that is not deleted, explore them
        for current_neighbour in neighbour_list:
            if self.is_white(current_neighbour) \
                    and (current_neighbour.get_first_id() + current_neighbour.get_target_decoy()) \
                    not in self.node_to_delete:
                self.dfs(current_neighbour, a_component)

        self.set_explored(start_node)

        # after the nodes are blackened,
        # add the protein or peptide into the a_component
        # with its neighbours
        # but get the object that is key, instead of the value
        # through the dict accession_object

        component_node = self.accession_object[start_node.get_first_id()]

        # TODO test
        # it is in in
        if start_node.get_first_id() == 'sp|Q9UI08-2|EVL_HUMAN':
            print(component_node.get_id())

        if start_node.get_first_id() == 'sp|Q9UI08-5|EVL_HUMAN':
            print(component_node.get_id())

        if isinstance(component_node, Protein):
            a_component.add_protein(component_node, neighbour_list)
        elif isinstance(component_node, Peptide):
            a_component.add_peptide(component_node, neighbour_list)



from typing import Dict, List, Optional, Tuple, Any
from node import Protein, Peptide
import random


class Component:
    _protein_dict: Dict[Protein, List[Peptide]]
    _peptide_dict: Dict[Peptide, List[Protein]]
    _covered_peptide: Dict[str, str]
    _target_dict: Dict[str, str]

    def __init__(self):
        self._protein_dict = {}
        self._peptide_dict = {}
        self._covered_peptide = {}
        self._target_dict = {}

    def add_peptide(self, current_peptide: Peptide,
                    neighbours) -> None:

        self._peptide_dict[current_peptide] = neighbours

    def add_protein(self, current_protein: Protein,
                    neighbours) -> None:

        self._protein_dict[current_protein] = neighbours

    def make_protein_list(self) -> \
            List[Tuple[List[str], float, str, Any]]:

        component_min_pro_list = []
        claimed_peptide = {}

        if len(self._protein_dict) == 0:
            return []

        while not self.all_component_peptides_covered():

            # find unselected protein with the most edges to uncovered peptide
            currently_selected_protein = self.find_most_uncovered_protein()

            # this should not happen, but sometimes it returns none
            if currently_selected_protein is None:
                break

            # set as selected, so subset is unselected ones
            currently_selected_protein.set_selected()

            # do 3 things, and check 2 things
            # 1. do set uncovered peptide as covered
            # 2. do find max score among uncovered peptide
            # 3. do map those uncovered peptide to the protein
            # 1. check that protein and peptide is same type
            # 2. check that max score after is not negative infinity
            max_score = float('-inf')
            could_cover_peptide = self._protein_dict[currently_selected_protein]

            for current_peptide in could_cover_peptide:
                # decoy should only match up with decoy, same for target
                same_type = currently_selected_protein.get_target_decoy() == current_peptide.get_target_decoy()
                assert same_type

                not_covered = current_peptide.get_first_id() not in self._covered_peptide
                if not_covered:
                    current_score = current_peptide.get_score()
                    if current_score > max_score:
                        max_score = current_score
                    self._covered_peptide[current_peptide.get_first_id()] = ''
                    # TODO: I do have extend now, putting all peptide to the
                    #  same list, but if we need to have peptide group in osw, then
                    #  need to change to append
                    claimed_peptide.setdefault(
                        currently_selected_protein.get_first_id(), []).extend(
                            current_peptide.get_id()
                        )

            assert max_score != float('-inf')

            currently_selected_protein.set_score(max_score)
            component_min_pro_list.append(currently_selected_protein)


        component_accession_list = []
        # component_min_pro_list is a list of protein that are selected
        # each protein contain their own accession list (list of str)
        # and score
        # claimed peptide is a dict of protein id to list of peptide id
        for current_protein in component_min_pro_list:
            component_accession_list.append(
                (
                    current_protein.get_id(), current_protein.get_score(),
                    current_protein.get_target_decoy(),
                    claimed_peptide[current_protein.get_first_id()]
                )
            )

        return component_accession_list

    def find_most_uncovered_protein(self) -> Optional[Protein]:
        """
        find an unselected protein that links to the most uncovered peptides
        :return:
        """

        most_edges_protein = random.choice(list(self._protein_dict.keys()))

        # record its number of edges
        most_uncovered = self.find_num_uncovered_peptides(most_edges_protein)

        # loop over the proteins of a_component
        for current_protein in self._protein_dict.keys():

            # of all the protein that are not selected, find the one with
            # the most edges (to peptides)
            if current_protein.is_selected():
                continue
            else:
                current_protein_num_uncovered = \
                    self.find_num_uncovered_peptides(current_protein)
                if current_protein_num_uncovered > most_uncovered:
                    most_uncovered = current_protein_num_uncovered
                    most_edges_protein = current_protein

        if most_uncovered == 0:
            return None


        return most_edges_protein

    def find_num_uncovered_peptides(self, current_protein: Protein) -> int:
        num_uncovered_peptides = 0
        peptide_neighbours = self._protein_dict[current_protein]
        for current_peptide in peptide_neighbours:
            # if it is covered or if it not same type, skip
            not_covered = current_peptide.get_first_id() not in self._covered_peptide
            same_type = current_peptide.get_target_decoy() == current_protein.get_target_decoy()
            if not_covered and same_type:
                num_uncovered_peptides += 1
            else:
                continue
        return num_uncovered_peptides

    def all_component_peptides_covered(self) -> bool:
        all_covered = True

        # find if any peptide is not yet covered
        for peptide in self._peptide_dict:
            if peptide.get_first_id() not in self._covered_peptide:
                all_covered = False
                break

        return all_covered



from __future__ import annotations

from typing import Any, Dict, List, Tuple

import sqlite3
import sys
# import igraph
from graph import Graph
from components import Component
from node import Peptide, Protein

"""main"""


def main(input_file: str, q_limit_pep: str, context: str, run_id: str) -> None:
    assert context == 'global' or context == 'run-specific'

    con = begin_connection(input_file)

    run_id = int(run_id)

    q_limit_pep = int(q_limit_pep)

    protein_peptide_graph = Graph()

    initialize(protein_peptide_graph, con, q_limit_pep, context, run_id)

    collapse(protein_peptide_graph)

    # visualize(protein_peptide_graph)

    component_list = separate(protein_peptide_graph)

    reduce(component_list, con)

    end_connection(con)


"""the helper function for main"""


def initialize(protein_peptide_graph: Graph, con,
               q_limit: int, context: str, run_id: int) -> None:
    peptide_id_list = get_all_peptide(con, q_limit, context, run_id)
    print("got all peptide")

    protein_id_list = get_all_protein_id(con, context, run_id)
    print("got all protein")

    linked_peptide_dict = get_all_link_for_protein(con, context, q_limit,
                                                   run_id)
    print("got all link from protein to peptide")

    protein_accession_dict = get_all_protein_accession(con)
    print("got all accession per protein sqlite id")

    linked_protein_dict = get_all_link_for_peptide(con)
    print("got all link from peptide to protein")

    protein_peptide_graph.add_peptide(peptide_id_list)
    print("added all peptide")

    protein_peptide_graph.add_protein(protein_id_list, linked_peptide_dict,
                                      protein_accession_dict)
    print("added all protein with edges")

    protein_peptide_graph.make_edges_from_peptide(linked_protein_dict,
                                                  protein_accession_dict)
    print("added edges from peptide")

    protein_peptide_graph.get_sort_keys()
    print("keys sorted by their id")

    # TODO test
    print("after just initialization")
    for key, value_list in protein_peptide_graph.node_dict.items():
        if key.get_first_id() == 'sp|Q96Q89-3|KI20B_HUMAN':
            value_accession_list = []
            for value in value_list:
                accession = value.get_first_id()
                value_accession_list.append(accession)
            print(value_accession_list)

    print("initialized")

    num_t_protein = 0
    num_d_peptide = 0
    num_d_protein = 0
    num_t_peptide = 0
    num_edges = 0
    for node in protein_peptide_graph.node_dict:
        num_edges += len(protein_peptide_graph.node_dict[node])
        if isinstance(node, Protein) and node.get_target_decoy() == '0':
            num_t_protein += 1
        elif isinstance(node, Peptide) and node.get_target_decoy() == '0':
            num_t_peptide += 1
        elif isinstance(node, Protein) and node.get_target_decoy() == '1':
            num_d_protein += 1
        elif isinstance(node, Peptide) and node.get_target_decoy() == '1':
            num_d_peptide += 1

    print("total nodes", len(protein_peptide_graph.get_node_dict()))
    print("total edges", num_edges)
    print("target protein", num_t_protein)
    print("target peptide", num_t_peptide)
    print("decoy protein", num_d_protein)
    print("decoy peptide", num_d_peptide)


def collapse(protein_peptide_graph: Graph) -> None:
    # protein_peptide_graph.collapse_graph_old()
    protein_peptide_graph.collapse_graph()


# def visualize(protein_peptide_graph):
#     pass
#     my_graph = protein_peptide_graph
#
#     their_graph = igraph.Graph()
#
#     # my_graph is a object with 1 dict
#     # one for protein, one for peptide
#     # the protein dict has protein as key
#     # and peptide as values
#
#     my_node_dict = my_graph.get_node_dict()
#
#     for vertex in my_node_dict:
#         if vertex.get_first_id() not in my_graph.node_to_delete:
#             my_string = ', '.join(map(str, vertex.get_id()))
#             their_graph.add_vertex(my_string)
#
#     print("vertex done")
#
#     protein_number = 1
#     peptide_number = 1
#     for key, value in my_node_dict.items():
#         if (key.get_first_id() not in my_graph.node_to_delete) \
#                 and isinstance(key, Protein):
#             vertex_string_1 = ', '.join(map(str, key.get_id()))
#             print("protein number", protein_number, "got protein vertex")
#             for element in value:
#                 vertex_string_2 = ', '.join(map(str, element.get_id()))
#                 their_graph.add_edge(vertex_string_1, vertex_string_2)
#                 peptide_number += 1
#             protein_number += 1
#
#     print("edges done")
#
#     layout = their_graph.layout_auto()
#     print("layout done")
#     plot = igraph.plot(their_graph, vertex_label=their_graph.vs["name"],
#                        layout=layout)
#     print("plot done")
#     plot.show()


def separate(protein_peptide_graph: Graph) -> List[Component]:
    # for all white peptide nodes, explore them
    protein_peptide_graph.make_accession_object_dict()

    if 'sp|Q9UI08-5|EVL_HUMAN' + '0' in protein_peptide_graph.node_to_delete:
        print('-5 it is deleted')
    else:
        print('-5 is not deleted')

    component_list = []

    for node in protein_peptide_graph.node_dict.keys():

        # if it is white and not deleted
        if protein_peptide_graph.is_white(node) and \
                (node.get_first_id() + node.get_target_decoy()) \
                not in protein_peptide_graph.node_to_delete:

            a_component = Component()
            protein_peptide_graph.dfs(node, a_component)
            component_list.append(a_component)

    print("separated")
    return component_list


# def visualize_component(component):
#     """
#     all nodes in a component are assumed to be not deleted
#     """
#     pass
#
#     their_graph = igraph.Graph()
#
#     # my_graph is a object with 1 dict
#     # one for protein, one for peptide
#     # the protein dict has protein as key
#     # and peptide as values
#
#     for vertex in component._protein_dict:
#         my_string = ', '.join(map(str, vertex.get_id()))
#         their_graph.add_vertex(my_string, type=False)
#     print("protein done")
#
#     for vertex in component._peptide_dict:
#         my_string = ', '.join(map(str, vertex.get_id()))
#         their_graph.add_vertex(my_string, type=True)
#     print("protein done")
#
#     protein_number = 1
#     peptide_number = 1
#     for key, value in component._protein_dict.items():
#         vertex_string_1 = ', '.join(map(str, key.get_id()))
#
#         for element in value:
#
#             vertex_string_2 = ', '.join(map(str, element.get_id()))
#             their_graph.add_edge(vertex_string_1, vertex_string_2)
#             peptide_number += 1
#         print("protein number", protein_number, "got all edges")
#         protein_number += 1
#
#     print("edges done")
#
#     layout = their_graph.layout_bipartite()
#     print("layout done")
#     plot = igraph.plot(their_graph, vertex_label=their_graph.vs["name"],
#                        layout=layout)
#     print("plot done")


def reduce(component_list: List[Component], con) -> None:
    min_pro_list = []

    for component in component_list:

        # this is a list of list of protein accession
        # sublist is one meta-protein vertex, contain multiple protein accession
        component_accession_list = component.make_protein_list()

        if len(component_accession_list) == 0:
            continue

        # this just a list of list of list of protein accession
        # sublist of it all belongs to the same component
        # sublist of the sublist belong to the same meta-protein vertex
        min_pro_list.append(component_accession_list)

    create_table_protein_group(con)
    create_table_protein_group_peptide_mapping(con)
    protein_group_data_entry(con, min_pro_list)

    print("reduced")


"""function that uses the Sqlite file directly"""


def begin_connection(db_name: str):
    con = sqlite3.connect(db_name)
    return con


def get_all_protein_id(con, context: str, run_id: int) -> List[Tuple[str, int]]:
    """
    choose all protein from the context
    :param run_id: the run id if the context is 'run-specific'
    :param context: the context, either global or 'run-specific'
    :param con: the connection to the sqlite database
    :return: a list of all proteins
    """
    c = con.cursor()
    # TODO: there must be a better solution
    if context == 'global':
        c.execute(
            """SELECT PROTEIN_ID, DECOY
            FROM SCORE_PROTEIN 
            INNER JOIN PROTEIN ON SCORE_PROTEIN.PROTEIN_ID = PROTEIN.ID
            WHERE CONTEXT='global'""")
    elif context == 'run-specific':
        c.execute(
            """SELECT PROTEIN_ID, DECOY
            FROM SCORE_PROTEIN 
            INNER JOIN PROTEIN ON SCORE_PROTEIN.PROTEIN_ID = PROTEIN.ID
            WHERE CONTEXT='run-specific' AND RUN_ID=:run_id""",
            {"run_id": run_id})
    all_protein_id_list = []
    for row in c.fetchall():
        all_protein_id_list.append((str(row[0]), row[1]))
    c.close()
    return all_protein_id_list


def get_all_peptide(con, q_limit: int, context: str, run_id: int) \
        -> List[Tuple[str, float, int]]:
    c = con.cursor()
    # choose all peptide from the global context and less than the threshold
    # TODO: there should be a more clever solution
    if context == 'global':
        c.execute(
            """SELECT PEPTIDE_ID, SCORE, DECOY
            FROM SCORE_PEPTIDE
            INNER JOIN PEPTIDE ON PEPTIDE.ID = SCORE_PEPTIDE.PEPTIDE_ID
            WHERE QVALUE<:q_limit AND SCORE_PEPTIDE.CONTEXT = 'global'""",
            {'q_limit': q_limit}
        )
    elif context == 'run-specific':
        c.execute(
            """SELECT PEPTIDE_ID, SCORE, DECOY
            FROM SCORE_PEPTIDE
            INNER JOIN PEPTIDE ON PEPTIDE.ID = SCORE_PEPTIDE.PEPTIDE_ID
            WHERE QVALUE<:q_limit AND SCORE_PEPTIDE.CONTEXT = 'run-specific' 
            AND RUN_ID=:run_id""",
            {'q_limit': q_limit, 'run_id': run_id}
        )
    all_peptide_id_list = []
    for row in c.fetchall():
        # each row is a tuple, since c.fetchall() returns a list of tuples
        all_peptide_id_list.append((str(row[0]), row[1], row[2]))

    c.close()
    return all_peptide_id_list

    # if I ever want to select only peptide or protein that fit a certain
    #  properties, use c.execute("SELECT * WHERE value=3 AND keyword=1")
    #  something like that, note this data is in SCORE_PROTEIN or SCORE_PEPTIDE


def get_all_link_for_protein(con, context, q_limit, run_id) -> Dict[
    str, List[Tuple[str, float, int]]]:
    c = con.cursor()
    if context == 'global':
        c.execute(
            """SELECT PEPTIDE_PROTEIN_MAPPING.PROTEIN_ID,
            PEPTIDE_PROTEIN_MAPPING.PEPTIDE_ID, SCORE_PEPTIDE.SCORE, 
            PEPTIDE.DECOY
            FROM PEPTIDE_PROTEIN_MAPPING
            INNER JOIN SCORE_PEPTIDE 
            ON SCORE_PEPTIDE.PEPTIDE_ID = PEPTIDE_PROTEIN_MAPPING.PEPTIDE_ID
            INNER JOIN PEPTIDE ON PEPTIDE.ID = SCORE_PEPTIDE.PEPTIDE_ID
            WHERE QVALUE<=:q_limit AND SCORE_PEPTIDE.CONTEXT = 'global'""",
            {'q_limit': q_limit})
    elif context == 'run-specific':
        c.execute(
            """SELECT PEPTIDE_PROTEIN_MAPPING.PROTEIN_ID,
            PEPTIDE_PROTEIN_MAPPING.PEPTIDE_ID, SCORE_PEPTIDE.SCORE, 
            PEPTIDE.DECOY
            FROM PEPTIDE_PROTEIN_MAPPING
            INNER JOIN SCORE_PEPTIDE 
            ON SCORE_PEPTIDE.PEPTIDE_ID = PEPTIDE_PROTEIN_MAPPING.PEPTIDE_ID
            INNER JOIN PEPTIDE ON PEPTIDE.ID = SCORE_PEPTIDE.PEPTIDE_ID
            WHERE QVALUE<=:q_limit AND SCORE_PEPTIDE.CONTEXT = 'run-specific' 
            AND RUN_ID=:run_id""",
            {'q_limit': q_limit, 'run_id': run_id})
    linked_peptide_dict = {}
    for row in c.fetchall():
        protein_id = str(row[0])
        peptide_id = str(row[1])
        peptide_score = row[2]
        peptide_decoy = row[3]
        peptide_info = (peptide_id, peptide_score, peptide_decoy)
        # if key was not in the dict, setdefault return the default value,
        # empty list here. if it was then it returns the value
        linked_peptide_dict.setdefault(protein_id, []).append(peptide_info)
    c.close()
    return linked_peptide_dict


def get_all_protein_accession(con) -> Dict[str, List[str]]:
    """
    returns a dictionary that pairs id with protein accession (as a list)
    """
    c = con.cursor()
    protein_accession_dict = {}
    c.execute(
        """SELECT ID, PROTEIN_ACCESSION 
            FROM PROTEIN"""
    )
    # if there are any row, split the row (which are text) into List[str]
    for row in c.fetchall():
        protein_sqlite_id = str(row[0])
        accession_sublist = row[1].split(",")

        stripped_accession_sublist = [s.strip() for s in accession_sublist]
        # if key was not in the dict, setdefault return the default value,
        # empty list here. if it was then it returns the value
        protein_accession_dict. \
            setdefault(protein_sqlite_id, []).extend(stripped_accession_sublist)

    c.close()

    for accession_list in protein_accession_dict.keys():
        protein_accession_dict[accession_list] \
            = list(set(protein_accession_dict[accession_list]))
    return protein_accession_dict


def get_all_link_for_peptide(con) -> Dict[str, List[Tuple[str, int]]]:
    """
    return a dict that serves as a mapping from peptide to protein
    """
    c = con.cursor()
    c.execute(
        """SELECT PEPTIDE_ID, PROTEIN_ID, PROTEIN.DECOY
        FROM PEPTIDE_PROTEIN_MAPPING
        INNER JOIN PROTEIN ON PROTEIN.ID = PEPTIDE_PROTEIN_MAPPING.PROTEIN_ID"""
    )
    linked_protein_dict = {}
    for row in c.fetchall():
        peptide_id = str(row[0])
        protein_id = str(row[1])
        protein_decoy = row[2]
        protein_info = (protein_id, protein_decoy)
        # if key was not in the dict, setdefault return the default value,
        # empty list here. if it was then it returns the value
        linked_protein_dict.setdefault(peptide_id, []).append(protein_info)
    c.close()
    return linked_protein_dict


def create_table_protein_group(con) -> None:
    """mpl = minimal protein list"""
    c = con.cursor()
    c.execute("""DROP TABLE IF EXISTS PROTEIN_GROUP""")
    c.execute("""CREATE TABLE PROTEIN_GROUP(
              COMPONENT_ID INTEGER,
              PROTEIN_GROUP_ID INTEGER, 
              PROTEIN_ID INTEGER,
              SCORE REAL,
              DECOY INT);""")
    con.commit()
    c.close()


def create_table_protein_group_peptide_mapping(con) -> None:
    c = con.cursor()
    c.execute("""DROP TABLE IF EXISTS PROTEIN_GROUP_PEPTIDE_MAPPING""")
    c.execute("""CREATE TABLE PROTEIN_GROUP_PEPTIDE_MAPPING(
              PROTEIN_GROUP_ID INTEGER, 
              PEPTIDE_ID INTEGER
              );""")
    con.commit()
    c.close()


def protein_group_data_entry(con, min_pro_list: List[List[Tuple[
    List[str], float, str, Any]]]) -> None:
    c = con.cursor()
    component_id = 0
    protein_group_id = 0
    for component in min_pro_list:
        for protein_group in component:
            protein_accession_list = protein_group[0]
            score = protein_group[1]
            decoy = protein_group[2]
            peptide_id_list = protein_group[3]
            for protein_accession in protein_accession_list:
                c.execute("""INSERT INTO PROTEIN_GROUP(COMPONENT_ID, 
                PROTEIN_GROUP_ID, PROTEIN_ID, SCORE, DECOY) VALUES(:component_id
                , :protein_group_id, :protein_accession, :score, :decoy)""",
                          {'component_id': component_id,
                           'protein_group_id': protein_group_id,
                           'protein_accession': protein_accession,
                           'score': score,
                           'decoy': decoy}
                          )
                con.commit()
            unique_peptide_id_list = list(set(peptide_id_list))
            for peptide_id in unique_peptide_id_list:
                c.execute("""INSERT INTO PROTEIN_GROUP_PEPTIDE_MAPPING(
                PROTEIN_GROUP_ID, PEPTIDE_ID) VALUES(:protein_group_id, :peptide_id)
                """, {'protein_group_id': protein_group_id,
                      'peptide_id': peptide_id})
                con.commit()
            protein_group_id += 1
        component_id += 1
    c.close()


def end_connection(con) -> None:
    con.close()


if __name__ == "__main__":
    if len(sys.argv) != 5:
        print("""usage: ParsimoniousProteinInferencer.py <sql file path>
              <q-value threshold for peptides> <context> <run id>""")
    # input_file path, q_limit_pep, context, run_id
    # if context is global, run_id can be anything
    # don't run global on 20180911_TIMS2_12-2_AnBr_SA_diaPASEF_Test10_42eV_1_A1_01_2927_test_lib_overridegroupid.osw
    # it does not have it
    main(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])



from __future__ import annotations

from abc import abstractmethod
from typing import List, Tuple


class Node:
    """=== Private Attributes ===

    _merge_number: same number means that these nodes are to be merged into
    a meta node, initially set as 0
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_id(self):
        pass

    @abstractmethod
    def add_id(self, accession_list):
        pass

    @abstractmethod
    def get_first_id(self):
        pass

    @abstractmethod
    def add_score(self, other_score: float):
        pass

    @abstractmethod
    def get_score(self):
        pass

    @abstractmethod
    def get_target_decoy(self):
        pass


class Protein(Node):
    _selected: bool
    _id_list: List[str]
    _first_id: str
    _sqlite_ids: List[str]
    _first_sqlite_id: str
    _score: float
    _decoy: int

    """initially, the protein group only has 1 accession number in it
    first accession serves as the immutable hash value for the dict key
    the protein_ids is just for making edges
    this is called protein, but it is actually referring to protein groups
    so initially it is a protein group with 1 protein"""

    def __init_subclass__(cls, **kwargs) -> None:
        pass

    def __init__(self, id_list: List[str], protein_sqlite_id: str, decoy: int) -> None:
        super().__init__()
        self._selected = False
        self._id_list = id_list
        self._first_id = id_list[0]
        self._first_sqlite_id = protein_sqlite_id
        self._sqlite_ids = [protein_sqlite_id]
        self._score = float('-inf')
        self._decoy = decoy

    def __hash__(self) -> int:
        return hash(self.get_first_id() + str(self._decoy))

    # bottleneck
    def get_both_first_accession(self, other: Protein) -> Tuple[str, str]:
        this_acc = self.get_first_id() + self.get_target_decoy()

        that_acc = other.get_first_id() + self.get_target_decoy()

        return this_acc, that_acc

    # bottleneck
    def __eq__(self, other: Protein) -> bool:
        this_acc, that_acc = self.get_both_first_accession(other)
        return this_acc == that_acc

    def __ne__(self, other: Protein) -> bool:
        this_acc, that_acc = self.get_both_first_accession(other)
        return this_acc != that_acc

    def __lt__(self, other: Protein) -> bool:
        this_acc, that_acc = self.get_both_first_accession(other)
        return this_acc < that_acc

    def __le__(self, other: Protein) -> bool:
        this_acc, that_acc = self.get_both_first_accession(other)
        return this_acc <= that_acc

    def __gt__(self, other: Protein) -> bool:
        this_acc, that_acc = self.get_both_first_accession(other)
        return this_acc > that_acc

    def __ge__(self, other: Protein) -> bool:
        this_acc, that_acc = self.get_both_first_accession(other)
        return this_acc >= that_acc

    def is_selected(self) -> bool:
        return self._selected

    def set_selected(self) -> None:
        self._selected = True

    # TODO: bottleneck
    def get_first_id(self) -> str:
        """
        this is for building the graph, because initially, there is only one id
        :return:
        """
        return self._first_id

    def add_id(self, accession_list: List[str]):
        self._id_list.extend(accession_list)

    def get_id(self) -> List[str]:
        return self._id_list

    def get_sqlite_id(self):
        return self._sqlite_ids

    def add_sqlite_id(self, protein_sqlite_id: str) -> None:
        self._sqlite_ids.append(protein_sqlite_id)

    def get_first_sqlite_id(self) -> str:
        return self._first_sqlite_id

    def set_score(self, score: float):
        self._score = score

    def add_score(self, other_score: float):
        """because protein group's score depends on the peptide it claims,
        so it is not changed during the merging of proteins """
        pass

    def get_score(self) -> float:
        return self._score

    def get_target_decoy(self) -> str:
        return str(self._decoy)


class Peptide(Node):
    _covered: bool
    _first_id: str
    _ids: List[str]
    _score: float
    _decoy: int

    """initially, the protein group only has 1 id in it"""

    def __init__(self, peptide_id_list: List[str], score: float,
                 decoy: int) -> None:
        """
        When initializing, the peptide_id_list only has 1 peptide id
        """
        super().__init__()
        self._covered = False
        self._ids = peptide_id_list
        self._first_id = peptide_id_list[0]
        self._score = score
        self._decoy = decoy

    def __hash__(self) -> int:
        return hash(self.get_first_id() + str(self._decoy))

    # bottleneck
    def get_both_ids(self, other: Peptide) -> Tuple[str, str]:
        this_id = self.get_first_id() + self.get_target_decoy()

        that_id = other.get_first_id() + self.get_target_decoy()

        return this_id, that_id

    # bottleneck
    def __eq__(self, other: Peptide) -> bool:
        this_id, that_id = self.get_both_ids(other)
        return this_id == that_id

    def __ne__(self, other: Peptide) -> bool:
        this_id, that_id = self.get_both_ids(other)
        return this_id != that_id

    def __lt__(self, other: Peptide) -> bool:
        this_id, that_id = self.get_both_ids(other)
        return this_id < that_id

    def __le__(self, other: Peptide) -> bool:
        this_id, that_id = self.get_both_ids(other)
        return this_id <= that_id

    def __gt__(self, other: Peptide) -> bool:
        this_id, that_id = self.get_both_ids(other)
        return this_id > that_id

    def __ge__(self, other: Peptide) -> bool:
        this_id, that_id = self.get_both_ids(other)
        return this_id >= that_id

    def is_covered(self) -> bool:
        return self._covered

    def set_covered(self) -> None:
        self._covered = True

    # bottleneck
    def get_first_id(self) -> str:
        return self._first_id

    def get_id(self) -> List[str]:
        return self._ids

    def add_id(self, id_list: List[str]):
        self._ids.extend(id_list)

    def add_score(self, other_score: float):
        """
        only keep the bigger score
        """
        self._score = max(self._score, other_score)

    def get_score(self) -> float:
        return self._score

    def get_target_decoy(self) -> str:
        return str(self._decoy)
